from ..helpers import eos
from ..helpers import alfaFunctions
from ..helpers.eosHelpers import A_fun, B_fun, getCubicCoefficients, getMixFugacity,getMixFugacityCoef
from ..solvers.cubicSolver import cubic_solver
from ..helpers import temperatureCorrelations as tempCorr
from ..helpers import mixing_rules

from numpy import log, exp, sqrt,absolute, array,sum
from scipy.optimize import fsolve, newton, root
from scipy.integrate import quad

def solve_eos(t,p,tc,pc,acentric,liq_compositions,vap_compositions,kij,method='pr',alfa_function='alfa_peng_robinson',mixing_rule='van_der_waals',diagram=False,properties=False,heat_capacity=None):
    # Vectorization
    tc = array(tc)
    pc= array(pc)
    acentric = array(acentric)
    liq_compositions=array(liq_compositions)
    vap_compositions = array(vap_compositions)
    kij = array(kij)
    
    # Method selection
    eos_fun = eos.selector(method)
    u,w,omega_a,omega_b,L = eos_fun()

    # Alfa function selection    
    alfa_fun = alfaFunctions.selector(alfa_function)
    alfa= alfa_fun(t,tc,acentric)
    
    Ai = A_fun(t,p,tc,pc,acentric,omega_a,alfa)
    Bi = B_fun(t,p,tc,pc,omega_b)

    # Mixing rules
    mixing_rule_used = mixing_rules.selector(mixing_rule)
    A_liq,B_liq,A_i_liq,Aij_liq,dAdT_liq = mixing_rule_used(liq_compositions,tc,acentric,kij,Ai,Bi,alfa,alfa_fun,t)
    A_vap,B_vap,A_i_vap,Aij_vap,dAdT_vap = mixing_rule_used(vap_compositions,tc,acentric,kij,Ai,Bi,alfa,alfa_fun,t)


    coefficients_liq = getCubicCoefficients(A_liq,B_liq,u,w)
    coefficients_vap = getCubicCoefficients(A_vap,B_vap,u,w)
   
    z_liq= cubic_solver(coefficients_liq,diagram,B_liq)
    z_vap = cubic_solver(coefficients_vap,diagram,B_vap)

   
    z_liq = z_liq[0] if isinstance(z_liq,tuple) else z_liq
    z_vap = z_vap[1] if isinstance(z_vap,tuple) else z_vap
    
    liq_fugacity_coef = getMixFugacityCoef(z_liq,A_liq,B_liq,A_i_liq,Bi,L)
    vap_fugacity_coef = getMixFugacityCoef(z_vap,A_vap,B_vap,A_i_vap,Bi,L)
    return (liq_fugacity_coef,vap_fugacity_coef)

def bubble_temperature(t,p,tc,pc,acentric,liq_compositions,vap_compositions,kij,delta_t=0.1,method='pr',alfa_function='alfa_peng_robinson',mixing_rule='van_der_waals'):
    liq_fugacity_coef,vap_fugacity_coef = solve_eos(t,p,tc,pc,acentric,liq_compositions,vap_compositions,kij,method,alfa_function,mixing_rule)
    Ki = liq_fugacity_coef/vap_fugacity_coef
    Sy = sum(Ki*liq_compositions)
    E = log(Sy)
    attempts=0
    new_t=t
    new_vap_compositions = vap_compositions

    while(absolute(E) >= 1e-9):
        if(attempts == 500):
            return 'Problem can not be solved'
        t0 = new_t + delta_t                    
        liq_fugacity_coef0,vap_fugacity_coef0 = solve_eos(t0,p,tc,pc,acentric,liq_compositions,new_vap_compositions,kij,method,alfa_function,mixing_rule)
        Ki0 =  liq_fugacity_coef0/vap_fugacity_coef0
        Sy0 = sum(Ki0*liq_compositions)
        E0 = log(Sy0)
        new_t = (new_t*t0*(E0-E))/(t0*E0-new_t*E)
        Sy = sum(Ki*liq_compositions)
        new_vap_compositions = (Ki*liq_compositions)/Sy
        liq_fugacity_coef,vap_fugacity_coef = solve_eos(new_t,p,tc,pc,acentric,liq_compositions,new_vap_compositions,kij,method,alfa_function,mixing_rule)
        Ki = liq_fugacity_coef/vap_fugacity_coef
        Sy = sum(Ki*liq_compositions)
        E=log(Sy)
        attempts +=1
    
    return(new_t,p,liq_compositions,new_vap_compositions)
        
def bubble_pressure(t,p,tc,pc,acentric,liq_compositions,vap_compositions,kij,delta_p=0.001,method='pr',alfa_function='alfa_peng_robinson',mixing_rule='van_der_waals'):
    liq_fugacity_coef,vap_fugacity_coef = solve_eos(t,p,tc,pc,acentric,liq_compositions,vap_compositions,kij,method,alfa_function,mixing_rule)
    Ki = liq_fugacity_coef/vap_fugacity_coef
    Sy = sum(Ki*liq_compositions)
    E = Sy -1
    attempts=0
    new_p=p
    new_vap_compositions = vap_compositions
  
    while(absolute(E) >= 1e-9):
        if(attempts == 100):
            return 'Probleam can not be solved'
        p0=new_p*(1+delta_p)
        liq_fugacity_coef0,vap_fugacity_coef0 = solve_eos(t,p0,tc,pc,acentric,liq_compositions,new_vap_compositions,kij,method,alfa_function,mixing_rule)
        Ki0 =  liq_fugacity_coef0/vap_fugacity_coef0
        Sy0 = sum(Ki0*liq_compositions)
        E0=Sy0-1
        new_p = (new_p*p0*(E0-E))/(p0*E0-new_p*E)
        Sy = sum(Ki*liq_compositions)
        new_vap_compositions = (Ki*liq_compositions)/Sy
        liq_fugacity_coef,vap_fugacity_coef = solve_eos(t,new_p,tc,pc,acentric,liq_compositions,new_vap_compositions,kij,method,alfa_function,mixing_rule)
        Ki = liq_fugacity_coef/vap_fugacity_coef
        Sy = sum(Ki*liq_compositions)
        E = Sy -1
        attempts +=1
    
    return(t,new_p,liq_compositions,new_vap_compositions)

def dew_temperature(t,p,tc,pc,acentric,liq_compositions,vap_compositions,kij,delta_t=0.1,method='pr',alfa_function='alfa_peng_robinson',mixing_rule='van_der_waals'):
    liq_fugacity_coef,vap_fugacity_coef = solve_eos(t,p,tc,pc,acentric,liq_compositions,vap_compositions,kij,method,alfa_function,mixing_rule)
    Ki = liq_fugacity_coef/vap_fugacity_coef
    Sx = sum(vap_compositions/Ki)
    E = log(Sx)
    attempts=0
    new_t=t
    new_liq_compositions = liq_compositions

    while(absolute(E) >= 1e-9):
        if(attempts == 500):
            return 'Probleam can not be solved'
        t0 = new_t + delta_t
        liq_fugacity_coef0,vap_fugacity_coef0 = solve_eos(t0,p,tc,pc,acentric,new_liq_compositions,vap_compositions,kij,method,alfa_function,mixing_rule)
        Ki0 =  liq_fugacity_coef0/vap_fugacity_coef0
        Sx0 = sum(vap_compositions/Ki0)
        E0 = log(Sx0)
        new_t = (new_t*t0*(E0-E))/(t0*E0-new_t*E)
        Sx = sum(vap_compositions/Ki)
        new_liq_compositions = vap_compositions/(Ki*Sx)
        liq_fugacity_coef,vap_fugacity_coef = solve_eos(new_t,p,tc,pc,acentric,new_liq_compositions,vap_compositions,kij,method,alfa_function,mixing_rule)
        Ki = liq_fugacity_coef/vap_fugacity_coef
        Sx = sum(vap_compositions/Ki)
        E = log(Sx)
        attempts +=1
    
    return(new_t,p,new_liq_compositions,vap_compositions)

def dew_pressure(t,p,tc,pc,acentric,liq_compositions,vap_compositions,kij,delta_p=0.001,method='pr',alfa_function='alfa_peng_robinson',mixing_rule='van_der_waals'):
    liq_fugacity_coef,vap_fugacity_coef = solve_eos(t,p,tc,pc,acentric,liq_compositions,vap_compositions,kij,method,alfa_function,mixing_rule)
    Ki = liq_fugacity_coef/vap_fugacity_coef
    Sx = sum(vap_compositions/Ki)
    E = Sx -1
    attempts=0
    new_p=p
    new_liq_compositions = liq_compositions
  
    while(absolute(E) >= 1e-9):
        if(attempts == 100):
            return 'Probleam can not be solved'
        p0=new_p*(1+delta_p)
        liq_fugacity_coef0,vap_fugacity_coef0 = solve_eos(t,p0,tc,pc,acentric,new_liq_compositions,vap_compositions,kij,method,alfa_function,mixing_rule)
        Ki0 =  liq_fugacity_coef0/vap_fugacity_coef0
        Sx0 = sum(vap_compositions/Ki0)
        E0=Sx0-1
        new_p = (new_p*p0*(E0-E))/(p0*E0-new_p*E)
        Sx = sum(vap_compositions/Ki)
        new_liq_compositions = vap_compositions/(Ki*Sx)
        liq_fugacity_coef,vap_fugacity_coef = solve_eos(t,new_p,tc,pc,acentric,new_liq_compositions,vap_compositions,kij,method,alfa_function,mixing_rule)
        Ki = liq_fugacity_coef/vap_fugacity_coef
        Sx = sum(vap_compositions/Ki)
        E = Sx -1 
        attempts +=1
          
    return(t,new_p,new_liq_compositions,vap_compositions)

def flash(t,p,tc,pc,acentric,feed_compositions,liq_compositions,vap_compositions,v_f,kij,delta_p=0.0001,method='pr',alfa_function='alfa_peng_robinson',mixing_rule='van_der_waals'):
    tau=1
    while(absolute(tau)> 1e-5):
        liq_fugacity_coef,vap_fugacity_coef = solve_eos(t,p,tc,pc,acentric,liq_compositions,vap_compositions,kij,method,alfa_function,mixing_rule)
        Ki = liq_fugacity_coef/vap_fugacity_coef
        S = sum((feed_compositions*(Ki-1))/(1+(v_f*(Ki-1))))
        S0 = sum((-feed_compositions*(Ki-1)**2)/(1+v_f*(Ki-1))**2)
        v_f = v_f-(S/S0)
        liq_compositions0 = feed_compositions/(1+v_f*(Ki-1))
        Sx=sum(liq_compositions0)
        liq_compositions = liq_compositions0/Sx
        vap_compositions0=liq_compositions0*Ki
        Sy=sum(vap_compositions0)
        vap_compositions=vap_compositions0/Sy
        tau=sum(absolute(liq_compositions*liq_fugacity_coef-vap_compositions*vap_fugacity_coef))
    
    return (t,p,feed_compositions,liq_compositions,vap_compositions,v_f)
    



from numpy import array,sum, sqrt
from scipy.misc import derivative

def selector(mixing_rule):
    if(mixing_rule == 'van_der_waals'):
        return van_der_waals
    else:
        return 'Mixing rule does not exist'

def van_der_waals(compositions,tc,acentric,kij,Ai,Bi,alfa,alfa_fun,T):
    shape = kij.shape
    B = sum(compositions*Bi)
    Aij=array([sqrt(Ai[i]*Ai[j])*(1-kij[i,j]) for i in range(0,len(Ai)) for j in range(0,len(Ai))]).reshape(shape)
    A = sum([sum(compositions[i]*compositions[j]*Aij[i,j]) for i in range(0,len(Ai)) for j in range(0,len(Ai))])
    A_i = array([(compositions[j]*Aij[:,j]) for j in range(0,len(Ai))])
    A_i =array([2* sum(A_i[:,i]) for i in range(0,len(Ai))])
    dAdT = sum([0.5*sum(compositions[i]*compositions[j]*Aij[i,j]*((T/alfa[i])*derivative(alfa_fun,T,args=(tc[i],acentric[i]))+(T/alfa[j])*derivative(alfa_fun,T,args=(tc[j],acentric[j])))) for i in range(0,len(Ai)) for j in range(0,len(Ai))])
    return (A,B,A_i,Aij,dAdT)



from .equilibrium import bubble_temperature,bubble_pressure, dew_temperature, dew_pressure
from numpy import max, array,arange, absolute, linspace
from matplotlib.pyplot import scatter, axhline, xlabel, ylabel, legend, plot,figure,show

def pressure_composition(t,tc,pc,acentric,kij,method='pr',alfa_function='alfa_peng_robinson',mixing_rule='van_der_waals',points=100):
    tc = array(tc)
    pc=array(pc)
    acentric=array(acentric)
    kij= array(kij)
    a =linspace(0,1,points)
    b=1-a
    liq_compositions=[]
    for i in range(0,len(a)):
        liq_compositions.append([a[i],b[i]])
    liq_compositions=array(liq_compositions)
    vap_compositions=liq_compositions

    aux1=bubble_pressure(t,1,tc,pc,acentric,liq_compositions[0],vap_compositions[0],kij,method=method,alfa_function=alfa_function,mixing_rule=mixing_rule)
    aux2=dew_pressure(t,1,tc,pc,acentric,liq_compositions[0],vap_compositions[0],kij,method=method,alfa_function=alfa_function,mixing_rule=mixing_rule)
    bubble_pressures=[aux1[1],]
    dew_pressures=[aux2[1],]
    
    for i in range(1,len(liq_compositions)):
        aux1=bubble_pressure(t,bubble_pressures[-1],tc,pc,acentric,liq_compositions[i],vap_compositions[i],kij,method=method,alfa_function=alfa_function,mixing_rule=mixing_rule)
        aux2=dew_pressure(t,dew_pressures[-1],tc,pc,acentric,liq_compositions[i],vap_compositions[i],kij,method=method,alfa_function=alfa_function,mixing_rule=mixing_rule)
        bubble_pressures.append(aux1[1])
        dew_pressures.append(aux2[1])

    figure(num=None, figsize=(8, 6), dpi=80, facecolor='w', edgecolor='k')
    plot(liq_compositions[:,:1],bubble_pressures,'b',label='Bubble line')
    plot(vap_compositions[:,:1],dew_pressures,'r',label='Dew line')
    xlabel('x,y')
    ylabel('Pressure [bar]')
    legend()
    

    
    

def temperature_composition(p,tc,pc,acentric,kij,method='pr',alfa_function='alfa_peng_robinson',mixing_rule='van_der_waals',points=100):
    tc = array(tc)
    pc=array(pc)
    acentric=array(acentric)
    kij= array(kij)
    a =linspace(0,1,points)
    b=1-a
    liq_compositions=[]
    for i in range(0,len(a)):
        liq_compositions.append([a[i],b[i]])
    liq_compositions=array(liq_compositions)
    vap_compositions=liq_compositions

    aux1=bubble_temperature(298,p,tc,pc,acentric,liq_compositions[0],vap_compositions[0],kij,method=method,alfa_function=alfa_function,mixing_rule=mixing_rule)
    aux2=dew_temperature(298,p,tc,pc,acentric,liq_compositions[0],vap_compositions[0],kij,method=method,alfa_function=alfa_function,mixing_rule=mixing_rule)
    bubble_temperatures=[aux1[0],]
    dew_temperatures=[aux2[0],]
    
    for i in range(1,len(liq_compositions)):
        aux1=bubble_temperature(bubble_temperatures[-1],p,tc,pc,acentric,liq_compositions[i],vap_compositions[i],kij,method=method,alfa_function=alfa_function,mixing_rule=mixing_rule)
        aux2=dew_temperature(dew_temperatures[-1],p,tc,pc,acentric,liq_compositions[i],vap_compositions[i],kij,method=method,alfa_function=alfa_function,mixing_rule=mixing_rule)
        bubble_temperatures.append(aux1[0])
        dew_temperatures.append(aux2[0])
      
    figure(num=None, figsize=(8, 6), dpi=80, facecolor='w', edgecolor='k')
    plot(liq_compositions[:,:1],bubble_temperatures,'b',label='Bubble line')
    plot(vap_compositions[:,:1],dew_temperatures,'r',label='Dew line')
    xlabel('x,y')
    ylabel('Temperature [K]')
    legend()



from numpy import sqrt, arccos, absolute, cos, pi

def cubic_solver_not_used(coefficients,diagram,B):
    alfa,beta,gamma = coefficients
    alfa= -1*alfa
    p=(3*beta-alfa**2)/3
    q=(27*gamma-9*alfa*beta+2*alfa**3)/27
    R=(p/3)**3+(q/2)**2

    if(R>0):
        first = -q*0.5+sqrt(R)
        second = -q*0.5-sqrt(R)
        
        if(second >0 and first >0 ):
            x = first**(1/3)+second**(1/3)-alfa/3
        elif(second>0 and first<0):
            x = -(-1*first)**(1/3) + second**(1/3)-alfa/3
        elif(second<0 and first>0):
            x = first**(1/3) -(-1*second)**(1/3)-alfa/3
        else:
            x = -(-1*first)**(1/3)-(-1*second)**(1/3)-alfa/3
        return x
        
    elif(R<0):
        theta = arccos(((-27/4)*(q**2/p**3))**(0.5))
        x1=-2*(absolute(q)/q)*(-p/3)**0.5 * cos(theta/3)-alfa/3
        x2=-2*(absolute(q)/q)*(-p/3)**0.5 * cos(theta/3+(2*pi/3))-alfa/3
        x3=-2*(absolute(q)/q)*(-p/3)**0.5 * cos(theta/3+(4*pi/3))-alfa/3
        x = [x1,x2,x3]
        return (min(x),max(x))
    else:
        raise Exception("cubic solver error")
    

def cubic_solver(coefficients,diagram,B):
    alfa,beta,gamma = coefficients
    C = 3*beta - alfa**2
    D = -alfa**3 + 4.5*alfa*beta-13.5*gamma
    Q = C **3 + D **2

    if(Q<= 0):
        theta = arccos(-D/(sqrt(-C**3)))
        
        z_liq = (alfa + 2* sqrt(-C)*cos(theta/3+(2*pi/3)))/3
        z_vap = (alfa + 2* sqrt(-C)*cos(theta/3))/3
        
        
        if (diagram):
            z_extra =  (alfa + 2* sqrt(-C)*cos(theta/3+(4*pi/3)))/3
            return(z_vap,z_liq,z_extra)
        
        if(z_liq < B):
            z_liq = (alfa + 2* sqrt(-C)*cos(theta/3))/3
        
        return (z_liq,z_vap)
        
    elif(Q>0):
        first = -D+sqrt(Q)
        second = -D-sqrt(Q)
        aux1=False
        aux2=False
        if(first<0):
            first *= -1
            aux1 = True
        if(second<0):
            second *= -1
            aux2 = True
            
            
        z = (alfa+(-1 if(aux1)else 1)*(first)**(1/3)+(-1 if(aux2) else 1)*(second)**(1/3))/3
        
        return z
    
    else:
        raise Exception("cubic solver error")
    
        


from numpy import sqrt, log

def selector(method):
    if(method=='pr'):
        cubic = peng_robinson
    elif(method=='rks'):
        cubic = redlich_kwong_soave
    else:
        return 'Method: ' + method + ' does not exist, define an allowed method'
    return cubic
    
def peng_robinson():
    u=2
    w=-1
    omega_a = 0.45723553
    omega_b = 0.077796074
    
    def L(z,B):
        return log((z+B*(1+sqrt(2)))/(z+B*(1-sqrt(2))))/(2*sqrt(2))
    
    return (u,w,omega_a,omega_b,L)


def redlich_kwong_soave():
    u=1
    w=0
    omega_a = 0.42748023
    omega_b = 0.08664035
    
    def L(z,B):
        return (log((z+B)/z))
    
    return (u,w,omega_a,omega_b,L)



from matplotlib.pyplot import scatter, axhline, xlabel, ylabel, legend, plot,figure,show
from mpl_toolkits.mplot3d import Axes3D
from scipy.interpolate import CubicSpline, interp1d
from numpy import linspace, log10, array, arange, append,zeros
from .equilibrium import solve_eos, solve_VLE
from mpl_toolkits.mplot3d import Axes3D
#import plotly
#import plotly.graph_objs as go

def pv(T,tc,pc,acentric,p_0=1,method='pr',alfa_function='alfa_peng_robinson',pvt=False):
    volumes = []
    pres = linspace(p_0,pc)
    R=83.14
    pressures=[]
    for p in pres:
        x=solve_eos(T,p,tc,pc,acentric,method=method,alfa_function=alfa_function,diagram=True)
        if(isinstance(x,tuple)):
            volumes.append(x[0]*R*T/p)
            volumes.append(x[1]*R*T/p)
            volumes.append(x[2]*R*T/p)
            pressures.append(p)
            pressures.append(p)
            pressures.append(p)
        else:
            volumes.append(x*R*T/p)
            pressures.append(p)
            
    if(pvt):
        return(array(pressures),array(volumes))

    scatter(log10(volumes),pressures)
    axhline(pc, color='k', linestyle='--')
    xlabel('log Volume [cm3]')
    ylabel('Pressure [bar]')
    legend(['Critical pressure','PV'])
    
    return 0 


def pt(tc,pc,acentric,p_0=1,t_0=298,method='pr',alfa_function='alfa_peng_robinson',pvt=False):
    R=83.14
    pressures = linspace(p_0,pc)
    t_first = solve_VLE(300,pressures[0],tc,pc,acentric,method=method,alfa_function=alfa_function,solving_for='temperature')
    temperatures=[t_first,]
    for i in range(1,len(pressures)):
        t = solve_VLE(temperatures[i-1],pressures[i],tc,pc,acentric,method=method,alfa_function=alfa_function,solving_for='temperature')
        temperatures.append(t)
    temperatures = array(temperatures)
    
    if(pvt):
        return (pressures,temperatures)
    
    plot(temperatures,pressures)
    axhline(pc, color='k', linestyle='--')
    xlabel('Temperature (K)')
    ylabel('Pressure [bar]')
    legend(['L-V','Critical pressure'])
    

def tv(P,tc,pc,acentric,t_0=298,method='pr',alfa_function='alfa_peng_robinson',pvt=False,temperatures_array=None, atomo4=False):
    volumes = []
    if(pvt):
        temp = temperatures_array
    else:
        temp = linspace(t_0,tc)
        
    R=83.14
    temperatures=[]
    for t in temp:
        x=solve_eos(t,P,tc,pc,acentric,method=method,alfa_function=alfa_function,diagram=True)
        if(isinstance(x,tuple)):
            volumes.append(x[0]*R*t/P)
            volumes.append(x[1]*R*t/P)
            volumes.append(x[2]*R*t/P)
            temperatures.append(t)
            temperatures.append(t)
            temperatures.append(t)
        else:
            volumes.append(x*R*t/P)
            temperatures.append(t)

    if(pvt):
        return (temperatures,volumes)
    
    scatter(log10(volumes),temperatures)
    axhline(tc, color='k', linestyle='--')
    xlabel('log Volume [cm3]')
    ylabel('Temperature [K]')
    legend(['Critical temperature','TV'])
    
    return 0 

def pvt(tc,pc,acentric,t_0=298,p_0=1,method='pr',alfa_function='alfa_peng_robinson'):
    temp = linspace(t_0,tc)
    general_pressures=array([])
    general_volumes=array([])
    general_temperatures=array([])
    i=1
    for T in temp:
        pressures,volumes = pv(T,tc,pc,acentric,p_0=1,method=method,alfa_function=alfa_function,pvt=True)
        aux=zeros(pressures.shape)
        aux.fill(T)
        general_volumes=append(general_volumes,volumes)
        general_pressures=append(general_pressures,pressures)
        general_temperatures=append(general_temperatures,aux)
        
    general_volumes=log10(general_volumes)
    
    #plotly.offline.init_notebook_mode()
    #fig = go.Figure(data=[go.Surface(z=general_pressures, x=general_volumes, y=general_temperatures)])
    #fig.update_layout(title='PVT',margin=dict(l=0, r=0, b=0, t=0))
    #plotly.offline.iplot(fig)
    
    fig = figure(figsize=[8,8])
    ax = fig.gca(projection='3d')

    ax.plot_trisurf(general_volumes, general_temperatures, general_pressures, linewidth=0.2, antialiased=True)
    ax.set_xlabel('log Volume [cm3]')
    ax.set_ylabel('Temperature [K]')
    ax.set_zlabel('Pressure [bar]')

    
    
    


from .equilibrium import bubble_temperature,bubble_pressure, dew_temperature, dew_pressure
from numpy import max, array,arange, absolute, linspace
from matplotlib.pyplot import scatter, axhline, xlabel, ylabel, legend, plot,figure,show



def pt(tc,pc,acentric,liq_compositions,vap_compositions,kij,method='pr',alfa_function='alfa_peng_robinson',mixing_rule='van_der_waals',points=100):
    tc = array(tc)
    pc=array(pc)
    max_pressure = max(pc)
    acentric=array(acentric)
    liq_compositions=array(liq_compositions)
    vap_compositions=array(vap_compositions)
    kij= array(kij)
    pressures = linspace(1,max_pressure,points)
    
    aux1=bubble_temperature(298,pressures[0],tc,pc,acentric,liq_compositions,vap_compositions,kij,method=method,alfa_function=alfa_function,mixing_rule=mixing_rule)
    aux2=dew_temperature(298,pressures[0],tc,pc,acentric,liq_compositions,vap_compositions,kij,method=method,alfa_function=alfa_function,mixing_rule=mixing_rule)
    bubble_temperatures=[aux1[0],]
    dew_temperatures=[aux2[0],]
    for i in range(1,len(pressures)):
        aux1 = bubble_temperature(aux1[0],pressures[i],tc,pc,acentric,liq_compositions,aux1[3],kij,method=method,alfa_function=alfa_function,mixing_rule=mixing_rule)
        aux2 = dew_temperature(aux2[0],pressures[i],tc,pc,acentric,aux2[2],vap_compositions,kij,method=method,alfa_function=alfa_function,mixing_rule=mixing_rule)
        
        if(absolute(aux1[0]-aux2[0]) <= 1 or absolute(aux1[0]-bubble_temperatures[-1])<=0.1  or absolute(aux2[0]-dew_temperatures[-1])<=0.1):
            break
        bubble_temperatures.append(aux1[0])
        dew_temperatures.append(aux2[0])

    #figure(num=None, figsize=(8, 6), dpi=80, facecolor='w', edgecolor='k')
    plot(bubble_temperatures,pressures[:len(bubble_temperatures)],'b',label='Bubble line')
    plot(dew_temperatures,pressures[:len(dew_temperatures)],'r',label='Dew line')
    xlabel('Temperature [K]')
    ylabel('Pressure [bar]')
    legend()



from numpy import sqrt, exp, where

def selector(alfa_function):
    if(alfa_function == 'soave'):
        fun = soave
    elif(alfa_function == 'alfa_peng_robinson'):
        fun=alfa_peng_robinson
    elif(alfa_function == 'pr78'):
        fun = pr78
    elif(alfa_function == 'mathias'):
        fun = mathias
    elif(alfa_function == 'stryjek_vera'):
        fun = stryjek_vera
    else:
        return 'Function: ' + alfa_function+ 'does not exist provide a valid function'
    return fun
    
def soave(t,tc,acentric):
    m = 0.48508 + 1.55171*acentric-0.15613*acentric**2
    reduced_temperature=t/tc
    return (1+m*(1-sqrt(reduced_temperature)))**2

def alfa_peng_robinson(t,tc,acentric):
    m=0.37464+1.54226*acentric-0.26992*acentric**2
    reduced_temperature=t/tc
    return (1+m*(1-sqrt(reduced_temperature)))**2

def pr78(t,tc,acentric):
    #if(acentric>0.49):
    #    m=0.379642+1.48503*acentric-0.164423*acentric**2+0.016666*acentric**3
    #else:
    #    m=0.37464+1.54226*acentric-0.26992*acentric**2
    m = where(acentric>0.49,0.379642+1.48503*acentric-0.164423*acentric**2+0.016666*acentric**3,0.37464+1.54226*acentric-0.26992*acentric**2)
    reduced_temperature=t/tc
    return (1+m*(1-sqrt(reduced_temperature)))**2


def mathias(t,tc,acentric,qi=0):
    reduced_temperature=t/tc
    m = 0.48508+1.55191*acentric-0.15613*acentric**2
    c = 1 + m/2 + 0.3*qi
    alfa = where(t>tc,exp(2*((c-1)/c)*(1-reduced_temperature**c)),(1+m*(1-reduced_temperature**(0.5))-qi*(1-reduced_temperature)*(0.7-reduced_temperature))**2)
        
    return alfa



def stryjek_vera(t,tc,acentric,k1):
    reduced_temperature=t/tc
    k0 = 0.378893+1.4897153*acentric-0.17131848*acentric**2+0.0196554*acentic**3
    
    if(t>tc):
        alfa = (1+k0*(1-sqrt(reduced_temperature)))**2
    else:
        alfa = (1+k0*(1-reduced_temperature**(0.5))+k1*(1-reduced_temperature)*(0.7-reduced_temperature))**2
        
    return alfa



from numpy import exp


    
def selector(number):
    if(number == '4'):
        return eq_4
    
    elif(number == '16'):
        return eq_16

    else:
        return None

def eq_4(t,constants):
    a,b,c,d = constants
    return (a+b*t+c*t**2+d*t**3)

def eq_16(t,constants):
    a,b,c,d,e = constants
    return (a+exp(b/t)+c+d*t+e*t**2)



from scipy.misc import derivative
from numpy import exp,log

def a_fun(t,tc,pc,acentric,omega_a,alfa):
    R = 83.14
    return omega_a*(R*tc)**2*alfa/pc

def b_fun(tc,pc,omega_b):
    R = 83.14
    return omega_b*(R*tc)/pc

def A_fun(t,p,tc,pc,acentric,omega_a,alfa):
    R = 83.14
    a = a_fun(t,tc,pc,acentric,omega_a,alfa)
    A = (a*p)/(R*t)**2
    return A

def dAdT_fun(t,p,tc,pc,acentric,omega_a,alfa):
    R=83.14
    dalfa=derivative(alfa,t,args=(tc,acentric))
    
    d_a = a_fun(t,tc,pc,acentric,omega_a,dalfa)
    return(t*d_a)*(p/(R*t)**2)

def B_fun(t,p,tc,pc,omega_b):
    R = 83.14
    b=b_fun(tc,pc,omega_b)
    B=(b*p)/(R*t)
    return B

def alfa_coefficient(B,u):
    return 1-(u-1)*B

def beta_coefficient(A,B,u,w):
    return A -u*B -u*B**2 +w*B**2

def gamma_coefficient(A,B,u,w):
    return A*B+w*B**2+w*B**3

def getCubicCoefficients(A,B,u,w):
    alfa = alfa_coefficient(B,u)
    beta = beta_coefficient(A,B,u,w)
    gamma = gamma_coefficient(A,B,u,w)
    return(alfa,beta,gamma)

def getPureFugacity(z,A,B,L,p):
    ln_fugacity_coef = -log(z-B) + (z-1) - A/B *L(z,B)
    fugacity = exp(ln_fugacity_coef)*p
    return fugacity

def getMixFugacity(z,A,B,A_i,Bi,L,compositions,p):
    ln_fugacity_coef = -log(z-B)+(z-1)*(Bi/B)+(A/B)*((Bi/B)-(A_i/A))*L(z,B)
    fugacity = exp(ln_fugacity_coef)*compositions*p
    return fugacity

def getMixFugacityCoef(z,A,B,A_i,Bi,L):
    ln_fugacity_coef = -log(z-B)+(z-1)*(Bi/B)+(A/B)*((Bi/B)-(A_i/A))*L(z,B)
    fugacity_coef = exp(ln_fugacity_coef)
    return fugacity_coef
    


from ..helpers import eos
from ..helpers import alfaFunctions
from ..helpers.eosHelpers import A_fun, B_fun, getCubicCoefficients, dAdT_fun, getPureFugacity
from ..solvers.cubicSolver import cubic_solver
from ..helpers import temperatureCorrelations as tempCorr
from numpy import log, exp, sqrt,absolute
from scipy.optimize import root
from scipy.integrate import quad
import json

def solve_eos(t,p,tc,pc,acentric,method='pr',alfa_function='alfa_peng_robinson',diagram=False,properties=False,heat_capacity=None):
    # Method selection
    eos_fun=eos.selector(method)
    u,w,omega_a,omega_b,L = eos_fun()

    # Alfa function selection    
    alfa_fun = alfaFunctions.selector(alfa_function)
    alfa= alfa_fun(t,tc,acentric)
    
    B = B_fun(t,p,tc,pc,omega_b)
    A = A_fun(t,p,tc,pc,acentric,omega_a,alfa)
    coefficients = getCubicCoefficients(A,B,u,w)
    

    x= cubic_solver(coefficients,diagram,B)
   
    if(diagram):
        return x
    
    if(isinstance(x,tuple)):
        z_liq = x[0]
        z_vap = x[1]
        
    else:
        z_liq=x
        z_vap=x
    
    liq_fugacity = getPureFugacity(z_liq,A,B,L,p)
    vap_fugacity = getPureFugacity(z_vap,A,B,L,p)
        
    if(properties):
        ideal_enthalpy = get_ideal_enthalpy(heat_capacity,t)/1000 # kmol to mol
        ideal_entropy = get_ideal_entropy(heat_capacity,t,p)/1000 #kmol to mol
        dAdt =  dAdT_fun(t,p,tc,pc,acentric,omega_a,alfa_fun)        
        enthalpy_liq = get_real_enthalpy(ideal_enthalpy,t,z_liq,A,dAdt,B,L)
        enthalpy_vap = get_real_enthalpy(ideal_enthalpy,t,z_vap,A,dAdt,B,L)
        entropy_liq = get_real_entropy(ideal_entropy,z_liq,A,dAdt,B,L)
        entropy_vap = get_real_entropy(ideal_entropy,z_vap,A,dAdt,B,L)

        response = {
            "liq_fugacity":liq_fugacity,
            "vap_fugacity":vap_fugacity,
            "enthalpy_liq":enthalpy_liq,
            "enthalpy_vap":enthalpy_vap,
            "entropy_liq":entropy_liq,
            "entropy_vap":entropy_vap,
            "z_liq":z_liq,
            "z_vap":z_vap
        }

        return response
    
         
    return (liq_fugacity, vap_fugacity)



def solve_VLE(t,p,tc,pc,acentric,solving_for='pressure',method='pr',alfa='alfa_peng_robinson'):
    if(solving_for=='pressure'):
        
        P = root(vle_pressure_objective_function,p,args=(t,tc,pc,acentric,method,alfa))
       
        return P.x[0]
    elif(solving_for=='temperature'):
        T = root(vle_temperature_objective_function,t,args=(p,tc,pc,acentric,method,alfa))
        
        return T.x[0]
        
def vle_pressure_objective_function(p,t,tc,pc,acentric,method='pr',alfa='alfa_peng_robinson'):
    
    liq_fugacity, vap_fugacity = solve_eos(t,p,tc,pc,acentric,method,diagram=False)
    
    return liq_fugacity-vap_fugacity

def vle_temperature_objective_function(t,p,tc,pc,acentric,method='pr',alfa='alfa_peng_robinson'):
    liq_fugacity, vap_fugacity = solve_eos(t,p,tc,pc,acentric,method,diagram=False)
    return liq_fugacity-vap_fugacity


def get_ideal_enthalpy(heat_capacity,t):
    number,constants = heat_capacity
    heat_capacity_equation =tempCorr.selector(number)
    enthalpy,_ = quad(heat_capacity_equation,298,t,args=(constants,))
    return enthalpy
    
    

def get_ideal_entropy(heat_capacity,t,p):
    R=8.314
    number,constants = heat_capacity
    heat_capacity_equation = lambda t,constants :tempCorr.selector(number)(t,constants)/t
    I,_ = quad(heat_capacity_equation,298,t,args=(constants,))
    entropy = I - R*log(p)
    return entropy


def get_real_enthalpy(ideal_enthalpy,t,z,A,dAdt,B,L):
    R=8.314
    enthalpy = ideal_enthalpy + R*t*(z-1+((dAdt-A)/B)*L(z,B))
    return enthalpy

def get_real_entropy(ideal_entropy,z,A,dAdt,B,L):
    R=8.314
    entropy = ideal_entropy + R*(log(z-B)+dAdt/B*L(z,B))
    return entropy


